#include <config.h>

!---------------------------------------------------
! setting for
! eigenvalue lower/upper bound estimation
!
! If USE_EIGEN_APPROX_FORBOUND is defined,
! eigensolver (Krylof-Schur method) is used
! with loose tolerance (10^-4). 
! This is rather slow for bound estimate.
!
! If USE_EIGEN_APPROX_FORBOUND is not defined
! Lanczos with safeguard method (by Zhou&Li)
! it used. This is faster and is recomended.
! But the bound is not sharp and can overestimate.
!---------------------------------------------------
#undef USE_EIGEN_APPROX_FORBOUND
!#define USE_EIGEN_APPROX_FORBOUND

!-----------------------------------------------------------------
! solver settings for multi-shift euqations
!-----------------------------------------------------------------
! if  _MULTI_SHIFT_CG is defined, we use multi-shift COCG algorithm is used.
! else normal CG algorithm is used seeuqntially.
! define _MULTI_SHIFT_CG is faster and recomended.
!------------------------------------------
!#define _MULTI_SHIFT_CG
#undef  _MULTI_SHIFT_CG

#ifndef _MULTI_SHIFT_CG
!------------------------------------------
! For normal CG algorithm the initial guesses are obtained via least-square 
! miminization. the ordering of the sequence affects the guess efficiency.
! with some experiments, defining _BACKWARD_ is recomended.
!------------------------------------------
!#undef _BACKWARD_
#define _BACKWARD_
#endif


module hmc_ratio_quark_wilson_class
!
!= Define wilson/clover quark Rational HMC algorithm (even/odd-site preconditioned)
!
!  Action = sum_{j=1,Nf} ye_{j}^+ R_n(Qee) ye_{j} - Nf TrLog[F]
!
! where
!
!  R_n(x) = (1/x)^(1/2)
!
!        v     v
!  Qee = Dee^+ Dee   : Hermitian operator
!  v
!  Dee     = 1ee - kapap^2 Fee \ Meo Foo \ Moe,   even/odd site preconditioned operator
!
!  v 
!  Dee^dag = 1ee - kapap^2 gamma5  Meo Foo \ Moe Fee \ gamma5
!                       v
!          = gamma5 Fee Dee Fee \ gamma5
!
!
! Rational Approximation : R_n(x) = P_n(x)/Q_n(x), x \in [a,b]
!
! Partial Fraction form is used.
!
!  R_n(x) = (1/x)^(1/2) =   alp_0 + \sum_{j=1..n} alp_j/(x-bet_j)
!
!  alp_j : residue
!  bet_j : pole
!
  use lattice_class
  use counter_class
  use timer_class
  use hmc_logfile_class
  use error_class
  use solver_parameter_class
  use metropolis_test_class
  use hmc_status_class
  use field_gauge_class
  use field_fermion_class
#ifdef _CLOVER
  use quark_clover_class
#else
  use quark_wilson_class
#endif
  use eigen_value_range_class
  implicit none
  private
  public :: hmc_ratio_quark_wilson
  public :: force
  public :: hamil
  public :: metropolis
  public :: reweighting
  public :: new
  public :: delete
  public :: read
  public :: save_config
  public :: print
  public :: print_statistics
  public :: RHMC_ALGORITHM
  public :: set_id
  public :: has_force_md_depth
  public :: has_global_metropolis_test
  public :: has_reweighting
  public :: get_eigen_Qee

  integer, parameter :: RHMC_ALGORITHM=4
  character(len=*), parameter :: SOLVER_LOG_FNAME_DEFAULT="wq_rhmc_solver_log."
  character(len=*), parameter ::  EIGEN_LOG_FNAME_DEFAULT="wq_rhmc_eigen_log."

  integer, parameter :: TYPE_SQUARE_ROOT    =+1
  integer, parameter :: TYPE_INV_SQUARE_ROOT=-1
  integer, parameter :: TYPE_QUAD_ROOT      =+2
  integer, parameter :: TYPE_INV_QUAD_ROOT  =-2

  type hmc_ratio_quark_wilson
    !
    ! HMC quark even/odd site preconditioned Rational HMC algorithm action
    !

#ifdef _CLOVER
    type(quark_clover),     public :: phys_param
    type(tfield_gluon_wg),  public, pointer :: XX => NULL()
#else
    type(quark_wilson),     public :: phys_param
#endif

    type(field_quark_wg),   private, allocatable :: y(:)

    type(solver_parameter), private :: hamil_inv  ! solver parameter for hamil
    type(solver_parameter), private :: force_inv  ! solver parameter for force

    type(hmc_logfile),      private :: solver_log
    type(hmc_logfile),      private :: eigen_log
    character(CHARLEN),     private :: solver_log_fname = SOLVER_LOG_FNAME_DEFAULT
    character(CHARLEN),     private :: eigen_log_fname  = EIGEN_LOG_FNAME_DEFAULT

    real(DP),               private, allocatable :: alp(:) ! residue of rational approximation
    real(DP),               private, allocatable :: bet(:) !    pole of rational approximation

    real(DP), private :: approx_range_factor_hamil = 1.01_DP  ! approximation range x in [a,b] :=>  x in [a/fac,a*fac]
    real(DP), private :: tol_approx_hamil          = EPSILON(1.0_DP)  ! rational approx for hamil (strict)

    type(eigen_value_range) :: eigen_force       ! rational approx range for force x \in [min,max]
    type(eigen_value_range) :: eigen_hamil       ! rational approx range for hamil x \in [min,max]


    type(eigen_value_range) :: eigen_hamil_hist(0:2)  ! eigen value range is saved. hist(0)=initail-hamil,hist(1)=last-hamil
    type(eigen_value_range) :: eigen_union            ! eigen value range union = U_{j=history} [min_j,max_j]

    type(eigen_value_range) :: eigen_func_eval        ! rational approx range for matrix function evaluation

    integer, private :: myid = 0             ! quark id
    integer, private :: nflavor = 0          ! number of flavor
    integer, private :: nratio_force = 0     ! rational order for force
    integer, private :: nratio = 0           ! rational order
    integer, private :: depth_hopping  = 1
    integer, private :: depth_clvtrlog = 1
    logical, private :: has_global_metropolis_test = .false.
    logical, private :: has_reweighting            = .false.
  end type

  interface new
    module procedure new_wilson_quark_rhmc
  end interface

  interface delete
    module procedure delete_wilson_quark_rhmc
  end interface

  interface read
    module procedure read_wilson_quark_rhmc
  end interface

  interface save_config
    module procedure save_config_hmc_ratio_quark_wilson
  end interface

  interface print
    module procedure print_wilson_quark_rhmc
  end interface

  interface print_statistics
    module procedure print_stat_wilson_quark_rhmc
  end interface

  interface force
    module procedure force_wilson_quark_rhmc
  end interface

  interface hamil
    module procedure hamil_wilson_quark_rhmc
  end interface

  interface assign_inv_mult_Qee_shift
#ifdef _GPU_SOLVER_
    module procedure v3_assign_inv_mult_Qee_shift
#else
    module procedure v1_assign_inv_mult_Qee_shift
#endif
  end interface

  interface metropolis
    !
    ! do global metropols test
    !
    module procedure metropolis_wilson_quark_rhmc
  end interface

  interface reweighting
    !
    ! compute reweighting factor
    !
    module procedure reweighting_wilson_quark_rhmc
  end interface

  interface set_id
    !
    ! set quark id
    !
    module procedure set_id_hrqw
  end interface

  interface has_force_md_depth
    !
    ! inquire whether this has force at the given md depth
    !
    module procedure has_force_wilson_quark_rhmc
  end interface

  interface has_global_metropolis_test
    !
    ! inquire whether this algorithm has global metropolis test
    !
    module procedure has_gmp_wilson_quark_rhmc
  end interface

  interface has_reweighting
    !
    ! inquire whether this algorithm has reweighting factor
    !
    module procedure has_reweighting_wilson_quark_rhmc
  end interface

contains 

subroutine set_id_hrqw(this,id)
!
! set quark # id
!
  implicit none
  type(hmc_ratio_quark_wilson), intent(inout) :: this
  integer, intent(in) :: id
  this%myid = id
  call set_id(this%phys_param,id)
  return
end subroutine

function has_force_wilson_quark_rhmc(this,md_depth) result(has_force)
!
! inquire whether this has force at the given md depth
!
  implicit none
  type(hmc_ratio_quark_wilson), intent(in) :: this
  integer, intent(in) :: md_depth
  logical :: has_force
  if ((md_depth == this%depth_hopping ) .or. &
 &    (md_depth == this%depth_clvtrlog)) then
    has_force = .TRUE.
  else
    has_force = .FALSE.
  endif
  return
end function

function has_gmp_wilson_quark_rhmc(this) result(has_gmp)
!
! inquire whether this algorithm has GMP
!
  implicit none
  type(hmc_ratio_quark_wilson), intent(in) :: this
  logical :: has_gmp
  has_gmp = this%has_global_metropolis_test
  return
end function

function has_reweighting_wilson_quark_rhmc(this) result(has_rew)
!
! inquire whether this algorithm has reweighting
!
  implicit none
  type(hmc_ratio_quark_wilson), intent(in) :: this
  logical :: has_rew
  has_rew = this%has_reweighting
  return
end function

subroutine new_wilson_quark_rhmc(this,id)
!
! initialize wilson/clover quark RHMC algorithm
!
  implicit none
  type(hmc_ratio_quark_wilson), intent(inout) :: this
  integer,                      intent(in)    :: id     ! quark id
  character(len=CHARLEN) :: str

  this%myid = id
  if (allocated(this%y)) deallocate(this%y)
  call new(this%phys_param,id)
  call new(this%hamil_inv)
  call new(this%force_inv)

  call new(this%eigen_force)
  call new(this%eigen_hamil)
  call new(this%eigen_hamil_hist(0))
  call new(this%eigen_hamil_hist(1))
  call new(this%eigen_hamil_hist(2))
  call new(this%eigen_union)

  write(str,'(I2.2)')id
  this%solver_log_fname = TRIM(this%solver_log_fname)//TRIM(str)
  this%eigen_log_fname  = TRIM(this%eigen_log_fname)//TRIM(str)

  call new(this%solver_log,this%solver_log_fname)
  call new(this%eigen_log,this%eigen_log_fname)

  return
end subroutine

subroutine delete_wilson_quark_rhmc(this)
!
! delete wilson/clover quark RHMC algorithm
!
  implicit none
  type(hmc_ratio_quark_wilson), intent(inout) :: this
  if (allocated(this%y)) deallocate(this%y)
  call delete(this%phys_param)
  call delete(this%solver_log)
  call delete(this%eigen_log)
  call delete(this%eigen_force)
  call delete(this%eigen_hamil)
  call delete(this%eigen_hamil_hist(0))
  call delete(this%eigen_hamil_hist(1))
  call delete(this%eigen_hamil_hist(2))
  call delete(this%eigen_union)
  return
end subroutine

subroutine read_wilson_quark_rhmc(this,iout)
!
! Read RHMC quark algorithm parameters and Initialize algorithm parameters.
!
  use comlib
  implicit none
  type(hmc_ratio_quark_wilson), intent(inout) :: this
  integer, intent(in) :: iout
  real(DP) :: rtmp0
  real(DP) :: eigen_min_force
  real(DP) :: eigen_max_force
  integer :: imb

  call new(this%hamil_inv)
  call new(this%force_inv)

  call read(this%phys_param,iout)
  if (nodeid==0) then
  read(iout,*)
  read(iout,*)this%hamil_inv%tol
  read(iout,*)this%force_inv%tol
  read(iout,*)this%depth_hopping
#ifdef _CLOVER
  read(iout,*)this%depth_clvtrlog
#endif
  read(iout,*)this%approx_range_factor_hamil
  read(iout,*)this%tol_approx_hamil
  read(iout,*)eigen_min_force
  read(iout,*)eigen_max_force
  read(iout,*)this%nratio_force
  endif
#ifndef _singlePU
  call comlib_bcast(this%hamil_inv%tol,0)
  call comlib_bcast(this%force_inv%tol,0)
  call comlib_bcast(this%depth_hopping,0)
  call comlib_bcast(this%depth_clvtrlog,0)
  call comlib_bcast(this%approx_range_factor_hamil,0)
  call comlib_bcast(this%tol_approx_hamil,0)
  call comlib_bcast(eigen_min_force,0)
  call comlib_bcast(eigen_max_force,0)
  call comlib_bcast(this%nratio_force,0)
#endif

  call set(this%eigen_force,eigen_min_force,eigen_max_force)

  this%nflavor = get_nflavor(this%phys_param)
  if (allocated(this%y)) deallocate(this%y)
  allocate(this%y(this%nflavor))  ! allocate multiple flavor

  return
end subroutine

subroutine save_config_hmc_ratio_quark_wilson(this,iout)
!
! Save Quark parameter on measurement configuration
!
  implicit none
  type(hmc_ratio_quark_wilson), intent(in) :: this
  integer, intent(in) :: iout
  call save_config(this%phys_param,iout)
  return
end subroutine

subroutine print_wilson_quark_rhmc(this)
!
! print out quark RHMC algorithm parameters on display
!
  implicit none
  type(hmc_ratio_quark_wilson), intent(inout) :: this
  integer :: imb

  if (nodeid==0) then

  write(*,'("**** Quark # ",I3," RHMC")')this%myid

  write(*,'("***** Physical parameters *****")')
  call print(this%phys_param)
  write(*,'("***** RHMC Algorithm parameters *****")')
  write(*,'(" Hamiltonian Stopping Condition :",E16.8)')this%hamil_inv%tol
  write(*,'("       Force Stopping Condition :",E16.8)')this%force_inv%tol
  write(*,'("  MD force depth PSfermion part :",I3)')this%depth_hopping
#ifdef _CLOVER
  write(*,'("  MD force depth CLV trlog part :",I3)')this%depth_clvtrlog
#endif
  write(*,'("Rational Approx Order for Force :",I4)') this%nratio_force
  write(*,'("Rational Approx Range for Force :")')
  write(*,'("             Qee in [a,b]=[",E24.15,",",E24.15,"]")') & 
 &                                           get_min(this%eigen_force), &
 &                                           get_max(this%eigen_force)

  write(*,'(80("-"))')

  endif

  return
end subroutine

subroutine print_stat_wilson_quark_rhmc(this)
!
! print out quark RHMC algorithm job statistics on display
!
  implicit none
  type(hmc_ratio_quark_wilson), intent(inout) :: this
  character(CHARLEN) :: name
  real(DP) :: hamil_ave,force_ave

  hamil_ave = get_average(this%hamil_inv)
  force_ave = get_average(this%force_inv)
  if (nodeid==0) then
  write(*,'("**** Quark # ",I3," RHMC")')this%myid

  call print(this%phys_param)

  write(*,'(" RHMC algorithm ")')
  write(*,'("         Hamil CG iter (averaged) :",F12.4)') hamil_ave
  write(*,'("         Force CG iter (averaged) :",F12.4)') force_ave
  write(*,'("       Minimum Eigenvalue for Qee :",E24.15)') get_min(this%eigen_union)
  write(*,'("       Maximum Eigenvalue for Qee :",E24.15)') get_max(this%eigen_union)

  endif

  return
end subroutine


subroutine force_wilson_quark_rhmc(this,BB,idepth,u)
!
! Calc MD force for Rational HMC action.
!
!   Action : Srhmc
!                        
!     Srhm  = \sum_{j=1,Nf} ye(j)^+ R_n(Qee) ye(j)  (even sites only)
!
!             - Nf*TrLog[F]
!            
! where
!
!  R_n(x) = (1/x)^(Nf/2) = alp_0 + \sum_{k=1,Nratio} alp_k/(x-bet_k)           
!           
!      v     v
!  Q = Dee^+ Dee
! 
! 
  use pfcoef_class
  implicit none
  type(hmc_ratio_quark_wilson), intent(inout) :: this    ! quark action/algorithm parameters
  integer,                      intent(in)    :: idepth  ! MD depth for recuresive MD integrator
  type(vfield_gluon_wg),        intent(in)    :: u       ! gauge field
  type(vfield_gluon_wg),        intent(inout) :: BB      ! pre-force contribution (\dot{u})

  real(DP) :: kappa2,fcoef, rtmp
  type(field_quark_wg),    allocatable :: fy,fx
  type(field_quark_eo_wg), allocatable :: fye(:),we
!test
  type(field_quark_eo_wg), allocatable :: vve(:)
!test
  type(hmc_status) :: status
  type(timer) :: etime
#ifdef _CLOVER
  real(DP) :: dummy
  real(DP) :: csw
  complex(DP) :: zickd8,ctmp
#endif
  type(partial_fraction_coef) :: pfcoef
  character(len=CHARLEN) :: str
  integer :: iout,iiter
  integer :: ir,nr
  integer :: iflv,nf
  real(DP) :: kappa,err

  call new(etime);call tic(etime)

#ifdef _CLOVER
  if (idepth == this%depth_hopping .or. &
 &    idepth == this%depth_clvtrlog) then
    !=======================================================
    ! comp. inverce clover term matrix
    ! fclinv = [1 - csw kappa/2 sigma_{mu,nu}F_{mu,nu}]^-1
    !=======================================================
    call make_inverse_clover_term(this%phys_param,0,dummy)
  endif
#endif

  if (idepth == this%depth_hopping) then
#ifdef _MD_DEPTH_
    if (nodeid==0) write(*,'("depth=",I3," : force quark poly hopping")')idepth
#endif

    write(str,'("# FORCE SOLVER")')
    call print(this%solver_log,TRIM(str))
    iout = get_file_unit(this%solver_log)

    nf    = get_nflavor(this%phys_param)
    kappa =   get_kappa(this%phys_param)
    nr    = this%nratio_force

    if (allocated(this%alp)) deallocate(this%alp)
    if (allocated(this%bet)) deallocate(this%bet)
    call new(pfcoef,TYPE_INV_SQUARE_ROOT,nr,get_min(this%eigen_force),get_max(this%eigen_force))
    allocate(this%alp(0:nr))
    allocate(this%bet(1:nr))
    this%alp(:) = pfcoef%alp(:)
    this%bet(:) = pfcoef%bet(:)

    call assign(this%eigen_func_eval,this%eigen_force)

    if (0 == nodeid) then
      write(iout,'("# order=",I3," minmaxerr=",E24.15," [a,b]=",2E24.15)')nr,pfcoef%minimaxerror,    &
 &                                                                        get_min(this%eigen_force), &
 &                                                                        get_max(this%eigen_force)
    endif

    call delete(pfcoef)

! test
    allocate(vve(1:nr))
    do ir=1,nr
      call new(vve(ir),0)
    enddo
! test
    allocate(fye(1:nr),fx,fy,we)
    do ir=1,nr
      call new(fye(ir),0)
    enddo
    call new(fx)
    call new(fy)
    call new(we,0)

    !================
    ! flavor loop
    !================
    do iflv = 1,nf

      !========================================
      !
      !  fye(j) = (Q-bet_j) \ ye(iflv)
      !
      !  using blocked/multi-shift solver
      !========================================
      err = this%force_inv%tol
      call assign_inv_mult_Qee_shift(this%phys_param,iout,err,iiter,  &
 &                                   fye,this%y(iflv)%eo(0),u,this%bet)
      call inc(this%force_inv,iiter)

      do ir=1,nr
        !========================================
        !           v
        !  fyo    = Moe fye(j)
        !           v
        !  fxe    = Dee fye(j)
        !
        !  fxe    = Fee \ fxe
        !                  v
        !  fxo    = gamma5 Moe gamma5 fxe
        !========================================
        call assign(fy%eo(0),fye(ir))
        call assign_mult_hop(      this%phys_param,fy%eo(1),fy%eo(0),u)
        call assign_mult_eoprec_wd(this%phys_param,fx%eo(0),fy%eo(0),u)

#ifdef _CLOVER
        call accum_mult_clover_term(fx%eo(0),this%phys_param)
#endif
        call assign_mult_gamma5(we,fx%eo(0))
        call assign_mult_hop(this%phys_param,fx%eo(1),we,u)
        call  accum_mult_gamma5(fx%eo(1))

        !======================================
        ! force from hopping matrix
        !======================================
        call copy_boundary(fx)
        call copy_boundary(fy)
        fcoef = (kappa**2)*this%alp(ir)
        call force_hmc_hopping(BB,fcoef,fx,fy)

#ifdef _CLOVER
        call assign_mult_hop(this%phys_param,fy%eo(0),fy%eo(1),u)
        !======================================
        ! force from clover term in hopping
        !======================================

          csw  = get_csw(this%phys_param)
        zickd8 = ((zi*csw*kappa**3)/8.0_DP)*this%alp(ir)
        call force_hmc_clover(this%XX,zickd8,fx,fy)
#endif
      enddo ! end of do ir

    enddo ! end of do iflv

    deallocate(fye,fx,fy,we)
    deallocate(this%alp)
    deallocate(this%bet)

  endif

#ifdef _CLOVER
  if (idepth == this%depth_clvtrlog) then
#ifdef _MD_DEPTH_
    if (nodeid==0) write(*,'("depth=",I3," : force quark poly trlog")')idepth
#endif
!======================================
! calc force from -Nf*log(det[F])
!======================================
    call force_clover_trlog(this%phys_param,this%XX)
  endif

  if (idepth == this%depth_hopping .or. &
 &    idepth == this%depth_clvtrlog) then
    call delete_inverse_clover_term(this%phys_param)
  endif

!======================================
! unlink clover term contribution
! XX is used for interfacing.
!======================================
  this%XX => NULL()
#endif 

  call toc(etime)
  if ((idepth == this%depth_hopping) .and. (idepth == this%depth_clvtrlog)) then
 
    write(str,'("force(RHMC,HOPPING,CLVTRLOG,",I3,"):")')idepth
    call print_log_timing(status,TRIM(str),etime)
    return
 
  else if (idepth == this%depth_hopping) then
 
    write(str,'("force(RHMC,HOPPING,",I3,"):")')idepth
    call print_log_timing(status,TRIM(str),etime)
    return
 
  else if (idepth == this%depth_clvtrlog) then
 
    write(str,'("force(RHMC,CLVTRLOG,",I3,"):")')idepth
    call print_log_timing(status,TRIM(str),etime)
    return
 
  endif

  return
end subroutine

subroutine hamil_wilson_quark_rhmc(this,ifi,u,action)
!
! Calc Hamil for Rational HMC action.
!
!   Action : Srhmc
!                        
!    Srhmc  = \sum_{j=1,Nf} ye(j)^+ R_n(Qee) ye(j)   (even sites only)
!
!             - Nf*TrLog[F]
!            
! where
!  R_n(x) = (1/x)^(1/2) = alp_0 + \sum_{j=1,Nratio} alp_j/(x-bet_j)
!        v     v
!  Qee = Dee^+ Dee
! 
! Initial Generation: 
! 
!  ye = G_n(Qee) eta
! 
!  eta : Gaussina noise 
! 
!  G_n(x) = (x)^(1/4)   : precise rational approximation for quad root.
! 
! Fianal computation
! 
!  Srhmc = |H_n(Qee)ye|^2,
! 
!  H_n(x) = (1/x)^(1/4) : precise rational approximation for inverse quad root.
! 
  use comlib
  implicit none
  type(hmc_ratio_quark_wilson), &
                         intent(inout) :: this   ! quark action/algorithm paramters
  integer,               intent(in)    :: ifi    ! target switch, 
                                                 ! * 0 for initial with generating PF field,
                                                 ! * 1 for final action,
                                                 ! * 2 for revcheck
  type(vfield_gluon_wg), intent(in)    :: u      ! gauge field
  real(DP),              intent(out)   :: action ! action value

  type(field_quark_eo_wg), allocatable :: we
  integer :: nf,iflv
  real(DP) :: Spf,Sclv,kappa
  character(CHARLEN) :: str
  type(hmc_status) :: status
  type(timer) :: etime
  integer :: iout,iiter
  real(DP) :: eval_min,eval_max,shift_max,shift_min,err,rtmp0,rtmp1,rtmp2
  logical :: do_get_eigen
  complex(DP) :: ctmp0

  call new(etime);call tic(etime)

  nf    = get_nflavor(this%phys_param)

  Spf =0.0_DP
  Sclv=0.0_DP
#ifdef _CLOVER
  !===========================================
  !  calc. F^-1 and
  !  actaion value of clover term trace log
  !===========================================
  call hamil_clover_trlog(this%phys_param,Sclv)
  if (0 == this%depth_clvtrlog) Sclv = 0.0_DP
#endif

  !========================================================
  ! Compute eigenvalue of Q to get
  !  approximation iterval [min,max]
  !========================================================
  call print(this%eigen_log,"%MIN and MAX EIGEN VALUES for Qee")
  iout  = get_file_unit(this%eigen_log)

  if ( (.not.is_null(this%eigen_hamil_hist(0))) .and. &
 &     (.not.is_null(this%eigen_hamil_hist(1))) .and. 0 == ifi) then
    !
    ! if previous traj is accepted, then
    ! the initial eigen is identical to the previous last eigen, else
    !                   is identical to the previous initial eigen.
    !
    if (get_previous_status(status)) then
      ! accepted case
      call assign(this%eigen_hamil,this%eigen_hamil_hist(1))
    else
      ! rejected case
      call assign(this%eigen_hamil,this%eigen_hamil_hist(0))
    endif
    eval_min = get_min(this%eigen_hamil)
    eval_max = get_max(this%eigen_hamil)
  else
#ifdef USE_EIGEN_APPROX_FORBOUND
    !------------------------------------------------------
    ! eigenvalue bound is derived with approximate
    ! max and min eigenvalues
    !------------------------------------------------------
    err   = 1.0e-4_DP
    iiter = 0

    if ( is_null(this%eigen_hamil_hist(0)) ) then
      shift_min = get_min(this%eigen_force)
      shift_max = get_max(this%eigen_force)
    else
      shift_min = get_min(this%eigen_hamil_hist(0))
      shift_max = get_max(this%eigen_hamil_hist(0))
    endif
    shift_min = shift_min/1.1_DP
    shift_max = shift_max*1.1_DP

    call get_eigen_Qee(this%phys_param,iout,err,iiter,u,ieo=0,eval=eval_min,shift=shift_min)
    call get_eigen_Qee(this%phys_param,iout,err,iiter,u,ieo=0,eval=eval_max,shift=shift_max)
#else
    !------------------------------------------------------
    ! eigenvalue bound is estimated using a simple Lanczos 
    ! safeguard method.
    !------------------------------------------------------
    call get_eigen_bound_Qee(this%phys_param,iout,u,ieo=0,eval_min=eval_min,eval_max=eval_max)
#endif
    call set(this%eigen_hamil,eval_min,eval_max)
  endif

  call assign(this%eigen_hamil_hist(ifi),this%eigen_hamil)    ! save for eval history
  call assign(this%eigen_func_eval,this%eigen_hamil)          ! for function evaluation

  this%eigen_union = union(this%eigen_union,this%eigen_hamil) ! to get largest eval range

  if (0 == nodeid) then
    write(iout,'("%EVAL_MIN_MAX:",I9,I2,2E18.8)')get_trajectory_number(status),ifi,eval_min,eval_max
  endif

!!==================
!  call check_rational_approx(this,iout,u)
!  stop
!!==================

  if ( (eval_min < get_min(this%eigen_force)) .or. &
 &     (get_max(this%eigen_force) < eval_max) ) then
    write(str,'(" Force approximation range does not contains the evaluated range.")')
    call warning_message(TRIM(str))
    write(str,'(" Evaluated range [a,b]=[",F12.4,",",F12.4,"]")')eval_min,eval_max
    call warning_message(TRIM(str))
    write(str,'("     Force range [a,b]=[",F12.4,",",F12.4,"]")')get_min(this%eigen_force), &
 &                                                               get_max(this%eigen_force)
    call warning_message(TRIM(str))
    write(str,'("Adjusting eigen_min_force and eigen_max_force is recomended.")')
    call warning_message(TRIM(str))
  endif

  write(str,'("# HAMIL SOLVER")')
  call print(this%solver_log,TRIM(str))
  iout = get_file_unit(this%solver_log)

  select case(ifi)
  case(0) ! INITIAL

    allocate(we)
    call new(we,0)

    Spf = 0.0_DP
    do iflv = 1,nf
      call set_gaussian_noise(we)

      !=============================
      ! calc. pseudo fermion action
      !=============================
      Spf = Spf + abs2(we)

      !============================
      !
      ! ye(iflv) = R_n(Q) we 
      ! 
      ! R_n(x) = (x)^(1/4)
      ! 
      ! with good approximation 
      ! 
      ! This needs |Q| range
      ! 
      !============================
      call assign_mult_function_Qee(this,TYPE_QUAD_ROOT,iout,this%y(iflv)%eo(0),we,u)

    enddo

    deallocate(we)

  case(1:2) ! FINAL : REVCHECK 

    allocate(we)
    call new(we,0)

    Spf = 0.0_DP
    do iflv = 1,nf
      !============================
      ! 
      ! we = R_n[Q] ye(iflv)
      !
      ! R_n[x] = (1/x)^(1/4)
      !
      ! with good approximation 
      ! 
      ! This needs |Q| range
      ! 
      !============================

      call assign_mult_function_Qee(this,TYPE_INV_QUAD_ROOT,iout,we,this%y(iflv)%eo(0),u)

      Spf = Spf + abs2(we)
    enddo
    deallocate(we)

  end select

  if (0 == this%depth_hopping) Spf = 0.0_DP

#ifdef _CLOVER
  call delete_inverse_clover_term(this%phys_param)
#endif

  action=Spf+Sclv

  write(str,'("@",I8,I2," Quark: ID:",I3," SQPF:",E24.16," SCLV:",E24.16)') &
 &  get_trajectory_number(status),ifi,this%myid,Spf,Sclv
  call print_log_action(status,TRIM(str))

  call toc(etime)
  select case(ifi)
  case(0)
    write(str,'("hamil0(RHMC):")')
  case(1:2)
    write(str,'("hamil1(RHMC):")')
  end select
  call print_log_timing(status,TRIM(str),etime)

  return
end subroutine

subroutine reweighting_wilson_quark_rhmc(this,u)
!
! Nothing is done
!
  implicit none
  type(hmc_ratio_quark_wilson), intent(inout) :: this ! quark action/algorihtm parameters
  type(vfield_gluon_wg),        intent(in)    :: u
  return
end subroutine

subroutine metropolis_wilson_quark_rhmc(this,u1,u0,ihit)
!
! Nothing is done
!
  implicit none
  type(hmc_ratio_quark_wilson), intent(inout) :: this ! quark action/algorihtm parameters
  type(vfield_gluon_wg),        intent(in)    :: u1   ! MD final gauge field
  type(vfield_gluon_wg),        intent(in)    :: u0   ! MD inital gauge field
  integer,                      intent(inout) :: ihit ! 1 for accept, 0 for reject
  return
end subroutine

subroutine v1_assign_inv_mult_Qee_shift(phys_param,iout,tol,iiter,Aye,ye,u,shift)
!                 
! Multiply inverse (Qee - shift_j) on a vector ye using multi-shift COCG algorithm.
!
!  Aye(j) = (Qee - shift_j) \ ye
!
! where
!        v       v
!  Qee = Dee^dag Dee
! 
! Refs. for multi-shift COCG algorithm;
! [1] R.Takayama, H.Hoshi, T.Sogabe, S.-L.Zhang, T.Fujiwara, Phys.Rev.B 73, 165108 (2006)
! [2] T.Sogabe, T.Hoshi, S.-L.Zhang, and T.Fujiwara, in: Frontiers of Computational Science,
!     eds. Y. Kaneda, H. Kawamura and M. Sasai, Springer-Verlag, Berlin/Heidelberg, 2007, pp.189-195
! 
  use solver_class
  use timer_class
  implicit none
#ifdef _CLOVER
  type(quark_clover), intent(inout) :: phys_param
#else
  type(quark_wilson), intent(inout) :: phys_param
#endif
  integer,  intent(in) :: iout
  real(DP), intent(in) :: tol
  integer,  intent(inout) :: iiter
  type(field_quark_eo_wg), intent(inout) :: Aye(:)     ! even/odd site fermion field block (output)
  type(field_quark_eo_wg), intent(in)    ::  ye        ! even/odd site fermion field (input)
  type(vfield_gluon_wg),   intent(in)    ::   u        ! gauge field
  real(DP), optional,      intent(in)    :: shift(:)   ! shift values

  type(field_quark_eo_wg), allocatable :: p(:),q,r,w
  type(counter) :: imult
  type(timer) :: solver_time
  character(len=CHARLEN) :: str
  integer  :: ir,nr
  integer  :: mult_count
  integer  :: iter
  logical :: is_shift
  complex(DP) :: ctmp
  logical  :: all_conved
  real(DP) :: etime,rtmp,bsrc
  real(DP) :: rho0,rho1,alp0
  real(DP), allocatable :: alp(:),bet(:),pim(:),pi0(:),pi1(:),err(:)
  logical,  allocatable :: conved(:)

  is_shift = present(shift)
  nr = SIZE(Aye(:))
  if (is_shift) then
  if (SIZE(shift(:)) /= nr) then
    write(str,'("Error: SIZE(shift) /= SIZE(Aye): v1_assign_inv_mult_Qee_shift:",A,I4)')__FILE__,__LINE__
    call error_stop(TRIM(str))
  endif
  endif

  call new(solver_time)
  call tic(solver_time)

  allocate(alp(1:nr),bet(1:nr),pi0(1:nr),pi1(1:nr),pim(1:nr),err(1:nr))
  allocate(conved(1:nr))
  allocate(p(1:nr),q,r)
  conved(:) = .FALSE.

  do ir=1,nr
    call new(Aye(ir),ye%ieo)
    call clear(Aye(ir))

    call new(p(ir),ye%ieo)
    call clear(p(ir))
  enddo
  call assign(r,ye)
  bsrc = SQRT(abs2(r))

  pim(:) = 1.0_DP
  pi0(:) = 1.0_DP
  alp(:) = 0.0_DP
  bet(:) = 0.0_DP
  alp0   = 1.0_DP

  imult = mult_iter

  rho0 = abs2(r)
  err(:) = SQRT(rho0)/bsrc
  !-----------
  ! p(1) = r
  !-----------
  call assign(p(1),r)

  !===============
  ! solver loop
  !===============
  do iter=1,5000

    !-------------------------------
    !       v
    !  q = (Qee - shift(1)) p(1) 
    !-------------------------------
    call assign_mult_Qee(phys_param,q,p(1),u)
    if (is_shift) then
      !========================================
      ! q = q + p(1) * (-shift(1))
      !========================================
      rtmp = -shift(1)
      call accum_add_mult(q,p(1),rtmp)
    endif

    !------------------------------
    ! alp(1) = rho0/<p(1)|A|p(1)>
    !------------------------------
    ctmp = prod(p(1),q)
    rtmp = REAL(ctmp,kind=DP)
    alp(1) = rho0/rtmp

    !----------------------------------------
    ! calc coefficients for shifted systems
    !----------------------------------------
    do ir=2,nr
      pi1(ir) = (1.0_DP+alp(1)*(-shift(ir)+shift(1)))*pi0(ir) &
 &                  + ((alp(1)*bet(1))/alp0)*(pi0(ir)-pim(ir))
      bet(ir) = bet(1)*(pim(ir)/pi0(ir))**2
      alp(ir) = alp(1)*(pi0(ir)/pi1(ir))

      if (.not.conved(ir)) then
        !------------------------------------
        ! p(j) = (1/pi0(j))*r + bet(j)*p(j)
        !------------------------------------
        call accum_mult(p(ir),bet(ir))
        rtmp = 1.0_DP/pi0(ir)
        call accum_add_mult(p(ir),r,rtmp)
      endif
   
      pim(ir) = pi0(ir)
      pi0(ir) = pi1(ir)
    enddo
    alp0 = alp(1)

    !---------------------------------
    ! x(j) = x(j) + alp(j) p(j)
    !---------------------------------
    do ir=1,nr
    if (.not.conved(ir)) then
      call accum_add_mult(Aye(ir),p(ir),alp(ir))
    endif
    enddo
    !---------------------------------
    ! r    = r    - alp(1) q
    !---------------------------------
    rtmp = - alp(1)
    call accum_add_mult(r,q,rtmp)

    err(1) = SQRT(abs2(r))/bsrc
#ifdef _DEBUG
    if (nodeid == 0) then
      do ir=1,nr
        if (ir==1) then 
          rtmp = err(1)
        else
          rtmp = err(1)/abs(pi0(ir))
        endif
        write(iout,'("@",I5," ERR_=",E24.16,I3)') iter,rtmp,ir
        write(*,'("@",I5," ERR_=",E24.16,I3)') iter,rtmp,ir
      enddo
    endif
#endif
    all_conved = .TRUE.
    do ir=2,nr
      rtmp = err(1)/abs(pi0(ir))
      if (rtmp < tol .and. (.not.conved(ir))) then
        conved(ir) = .TRUE.
        err(ir) = rtmp
      endif
      all_conved = (all_conved .AND. conved(ir))
    enddo
    if ( (err(1) < tol) .and. all_conved) then
      mult_count = get_count(mult_iter)-get_count(imult)
      if (0 == nodeid) then
        write(iout,'("# multi-shift COCG converged.")')
        do ir=1,nr
          if (ir < nr) then 
            write(iout,'("#",I5," ERR_=",E24.16,I3)')iter,err(ir),ir
          else
            write(iout,'("#",I5," ERR_=",E24.16,I3," MULT_=",I10)',advance='no') iter,err(ir),ir,mult_count
          endif
        enddo
      endif
      iiter = iter
      goto 1000
    endif

    !---------------------------------
    ! bet(1) = rho1/rho0
    !---------------------------------
    rho1 = abs2(r)
    bet(1) = rho1/rho0
    rho0 = rho1

    !---------------------------------
    ! p(1) = p(1)*bet(1) + r
    !---------------------------------
    call accum_mult_add(p(1),bet(1),r)

  enddo

  mult_count = get_count(mult_iter)-get_count(imult)
  if (nodeid == 0) then
    write(iout,'("# multi-shift COCG iteration does not converge.")')
    do ir=1,nr
      if (ir < nr) then 
        write(iout,'("#",I5," ERR_=",E24.16,I3)')iter,err(ir),ir
      else
        write(iout,'("#",I5," ERR_=",E24.16,I3," MULT_=",I10)') iter,err(ir),ir,mult_count
      endif
    enddo
  endif
  call error_stop("# multi-shift COCG solver stop.")

1000 continue

  call toc(solver_time)
  etime = get_elapse(solver_time)
  if (nodeid == 0) then
    write(iout,'(" ETIME_=",E24.16)')etime
  endif

#ifdef _DEBUG
  do ir=1,nr
    !---------------------------------
    ! q = Qee x(ir)
    !---------------------------------
    call assign_mult_Qee(phys_param,q,Aye(ir),u)
    if (is_shift) then
      !-----------------------------
      ! q = q + x(ir)*(-shift(ir))
      !-----------------------------
      rtmp = -shift(ir)
      call accum_add_mult(q,Aye(ir),rtmp)
    endif
    call assign_sub(r,ye,q)
    rtmp = SQRT(abs2(r))/bsrc
    if (nodeid == 0) then
      write(iout,'("X",I5," ERR_=",E24.16,I3)') iter,rtmp,ir
    endif
  enddo
#endif

  deallocate(alp,bet,pi0,pi1,pim,err)
  deallocate(conved)
  deallocate(p,q,r)

  return
end subroutine


subroutine v2_assign_inv_mult_Qee_shift(phys_param,iout,tol,iiter,Aye,ye,u,shift)
!                 
! Multiply inverse (Qee - shift_j) on a vector ye using normal CG algorithm.
!
!  Aye(j) = (Qee - shift_j) \ ye
!
! where
!        v       v
!  Qee = Dee^dag Dee
! 
! 
  use solver_class
  use timer_class
  implicit none
#ifdef _CLOVER
  type(quark_clover), intent(inout) :: phys_param
#else
  type(quark_wilson), intent(inout) :: phys_param
#endif
  integer,  intent(in) :: iout
  real(DP), intent(in) :: tol
  integer,  intent(inout) :: iiter
  type(field_quark_eo_wg), intent(inout) :: Aye(:)   ! even/odd site fermion field block (output)
  type(field_quark_eo_wg), intent(in)    ::  ye      ! even/odd site fermion field (input)
  type(vfield_gluon_wg),   intent(in)    ::   u      ! gauge field
  real(DP), optional,      intent(in)    :: shift(:) ! shift values

  type(field_quark_eo_wg), allocatable :: we
  type(cg_alg), allocatable :: solver
  type(counter) :: imult
  type(timer) :: solver_time
  character(len=CHARLEN) :: str
  integer  :: ir,nr
  integer  :: mult_count
  integer  :: istat
  logical :: is_shift
  real(DP) :: etime,rtmp

  is_shift = present(shift)
  nr = SIZE(Aye(:))
  if (is_shift) then
  if (SIZE(shift(:)) /= nr) then
    write(str,'("Error: SIZE(shift) /= SIZE(Aye): v2_assign_inv_mult_Qee_shift:",A,I4)')__FILE__,__LINE__
    call error_stop(TRIM(str))
  endif
  endif

  do ir=1,nr
    call new(Aye(ir),ye%ieo)
    call clear(Aye(ir))
  enddo
  allocate(we)
  call new(we,ye%ieo)

  allocate(solver)

#ifdef _BACKWARD_
  do ir=nr,1,-1
#else
  do ir=1,nr
#endif

    call new(solver_time)
    call tic(solver_time)

#ifdef _BACKWARD_
    if (ir == nr) then
#else
    if (ir == 1) then
#endif
      call new(solver,NSIZE=COL*SPIN*NTH*NZ*NY*NX,max_iter=5000,guess=0,tol=tol)
    else
      call new(solver,NSIZE=COL*SPIN*NTH*NZ*NY*NX,max_iter=5000,guess=1,tol=tol)
      !====================
      ! set initial guess
      !====================
      call get_initial_guess_multi_shift(phys_param,Aye(ir),ir,ye,Aye,shift)
      call pack(Aye(ir),solver%dst_vec)
    endif

    imult = mult_iter

    !===============
    ! set source
    !===============
    call pack(ye,solver%src_vec)

    !===============
    ! solver loop
    !===============
    do 
      call solve(solver)
      istat = get_status(solver)
      select case (istat)
      case (OP_NOP)
        cycle
      case (OP_DO_MATVEC)

        !==============================
        !            v
        ! Aye(ir) = [Qee - shift(ir)] we
        !==============================
        call unpack(solver%src_vec,we)
        call assign_mult_Qee(phys_param,Aye(ir),we,u)
        if (is_shift) then
          !========================================
          ! Aye(ir) = Aye(ir) + we * (-shift(ir))
          !========================================
          rtmp = -shift(ir)
          call accum_add_mult(Aye(ir),we,rtmp)
        endif
        call   pack(Aye(ir),solver%dst_vec)

      case (OP_PRINT_STATUS)
#ifdef _DEBUG
        if (nodeid == 0) then
          write(iout,'("@",I5," ERR_=",E24.16)') &
 &                            get_current_iteration(solver),&
 &                                get_residual_norm(solver)
        endif
#endif
      case (OP_CONVERGED)
        if (nodeid == 0) then
          mult_count = get_count(mult_iter)-get_count(imult)
#ifdef _DEBUG
          write(iout,'("@",I5," ERR_=",E24.16)') &
 &                            get_current_iteration(solver),&
 &                                get_residual_norm(solver)
#endif
          str = TRIM(get_name(solver))
          write(iout,'("#",A," iteration converged.")') TRIM(str)
          write(iout,'("#",I5," ERR_=",E24.16," MULT_=",I10)',advance='no')   &
 &                            get_current_iteration(solver),&
 &                                get_residual_norm(solver),mult_count

        endif
        iiter = get_current_iteration(solver)
        exit      
      case (OP_MAXITER_REACHED)
        if (nodeid == 0) then
          mult_count = get_count(mult_iter)-get_count(imult)
#ifdef _DEBUG
          write(iout,'("@",I5," ERR_=",E24.16)') &
 &                            get_current_iteration(solver),&
 &                                get_residual_norm(solver)
#endif
          str = TRIM(get_name(solver))
          write(iout,'("#",A," iteration does not converge.")') TRIM(str)
          write(iout,'("#",I5," ERR_=",E24.16," MULT_=",I10)')     &
 &                            get_current_iteration(solver),&
 &                                get_residual_norm(solver),mult_count
        endif
        call error_stop("Q^-1 solver did not converge.")
      case default
        call error_stop("something is wrong in Q^-1 solver.")
      end select
    enddo

    !===========================
    ! keep solutions Aye(j) 
    !===========================
    call unpack(solver%dst_vec,Aye(ir))
    call delete(solver)

    call toc(solver_time)
    etime = get_elapse(solver_time)
    if (nodeid == 0) then
      write(iout,'(" ETIME_=",E24.16)')etime
    endif

  enddo

  deallocate(solver,we)

  return
end subroutine


subroutine get_eigen_bound_Qee(phys_param,iout,u,ieo,eval_min,eval_max)
!                                v                 v
! Estimates eigenvalue bound for Qee  [eval_min < |Qee| < eval_max]
!
! Eigenvalue upper bound is estimabed by Lanczos algorithm.
! Ref.
! Yunkai Zhou and Len-Cang Li, "Bounding the spectrum of large Hermitian matrices",
! Linear Algebra and its Applications 435 (2011) 480-493.
!
!  phys_param : physics parameters
!        iout : log file unit
!           u : gauge link field
!         ieo : bound is estimated for [0 for even sites/ 1 for odd sites (Qee or Qoo)]
!    eval_min : lower bound of Qee/Qoo
!    eval_max : upper bound of Qee/Qoo
!
  use timer_class
  use f95_lapack, only : LA_HEEV
  implicit none
#ifdef _CLOVER
  type(quark_clover), intent(inout) :: phys_param
#else
  type(quark_wilson), intent(inout) :: phys_param
#endif
  integer,               intent(in)    :: iout    ! log file unit
  type(vfield_gluon_wg), intent(in)    :: u       ! gauge field
  integer,               intent(in)    :: ieo     ! even/odd site index
  real(DP),              intent(out)   :: eval_max,eval_min    ! eigen bound

  type(timer) :: solver_time
  type(field_quark_eo_wg), allocatable :: qq(:),qw(:)
  real(DP),    allocatable :: ev(:)
  complex(DP), allocatable :: TT(:,:),vv(:),AA(:,:)
  complex(DP) :: ctmp
  real(DP) :: rtmp, rtmp2, etime
  real(DP) :: dmax, dmin, evmax
  integer :: imax, imin, iev, istp
  integer :: i,j,k, jj(1), kmax, ierr
  integer, parameter ::  KDIM = 10
  integer :: NSTP,iiter
  real(DP), parameter :: tol_lcz = 1.0e-8_DP
  character(len=CHARLEN) :: str

  call new(solver_time)
  call tic(solver_time)

  NSTP = 1

  allocate(qq(1:KDIM+1),TT(1:KDIM+1,1:KDIM+1),AA(1:KDIM,1:KDIM),ev(1:KDIM),qw(1))
  do j=1,KDIM+1
    call new(qq(j),ieo)
  enddo
  call new(qw(1),ieo)

  do iev = 0,1
  !
  ! iev = 0 for non-shifted, iev = 1 for shifted
  !

    allocate(vv(COL*SPIN*NTH*NZ*NY*NX))
    vv(:) = cmplx(1.0_DP,1.0_DP)
    call unpack(vv,qq(1))
    deallocate(vv)

    !------------------------------------
    ! Enhance larger eigenmode
    ! with KDIM step power iteration
    !------------------------------------
    do istp = 1,KDIM
      select case(iev)
      case(0)
        call assign_mult_Qee(phys_param,qq(KDIM+1),qq(1),u)
        call assign(qq(1),qq(KDIM+1))
      case(1)
        iiter=0
        call assign_inv_mult_Qee_shift(phys_param,iout,tol_lcz,iiter,qw,qq(1),u)
        call assign(qq(1),qw(1))
      end select
      rtmp = abs2(qq(1))
#ifdef _DEBUG_
      if (nodeid==0) write(iout,'("# |q|=",E24.15)')sqrt(rtmp)
#endif
      rtmp = SQRT(1.0_DP/rtmp)
      call accum_mult(qq(1),rtmp)
    enddo

1000 continue

    do istp = 1,NSTP

      ev(:)   = Z0
      TT(:,:) = Z0
      AA(:,:) = Z0

      rtmp = abs2(qq(1))
      rtmp = SQRT(1.0_DP/rtmp)
      call accum_mult(qq(1),rtmp)
  
      !------------------------------------
      ! KDIM step Lanczos iteration
      !------------------------------------
      kmax = KDIM
      do j = 1,KDIM
        !-----------------------------------
        ! qq(j+1) = A qq(j)
        !-----------------------------------
        select case(iev)
        case(0)
          call assign_mult_Qee(phys_param,qq(j+1),qq(j),u)
        case(1)
          iiter=0
          call assign_inv_mult_Qee_shift(phys_param,iout,tol_lcz,iiter,qw,qq(j),u)
          call assign(qq(j+1),qw(1))
        end select

        !----------------------------------------
        ! qq(j+1) = qq(j+1) - sum_{k=1,j} qq(k) <qq(k)|qq(j+1)>
        !----------------------------------------
        do k=1,j
          ctmp = prod(qq(k),qq(j+1))
          TT(k,j) = ctmp
          call accum_add_mult(qq(j+1),qq(k),-ctmp)
        enddo
        do k=1,j
          ctmp = prod(qq(k),qq(j+1))
          TT(k,j) = TT(k,j) + ctmp
          call accum_add_mult(qq(j+1),qq(k),-ctmp)
        enddo
        rtmp = abs2(qq(j+1))
        rtmp2 = SQRT(rtmp)
        TT(j+1,j) = CMPLX(rtmp2,0.0_DP,kind=DP)

        if ( rtmp2 < tol_lcz ) then
          if (0 == nodeid) write(iout,'("# RHMC Lanczos breakdown. j=",I3)')j
          kmax = j
          exit
        endif

        !-------------------------------
        ! qq(j+1) = qq(j+1)/|qq(j+1)|
        !-------------------------------
        rtmp = SQRT(1.0_DP/rtmp)
        call accum_mult(qq(j+1),rtmp)
      enddo

      do k=1,kmax
      do j=1,kmax
        AA(j,k) = (TT(j,k) + CONJG(TT(k,j)))*0.5_DP
      enddo
      enddo
#ifdef _DEBUG_
      do k=1,kmax
      do j=1,kmax
        if (nodeid==0) write(iout,'("AA(",I3,",",I3,")=",2E24.15)')j,k,AA(j,k)
      enddo
      enddo
#endif

      TT(1:kmax,1:kmax) = AA(1:kmax,1:kmax)
      call LA_HEEV(AA(1:kmax,1:kmax),ev(1:kmax),JOBZ='V',info=ierr)
      if (ierr /= 0) then
        write(str,'(" get_eigen_bound_Qee: LA_HEEV error info=",I3)')ierr
        call error_stop(TRIM(str))
      endif

      !----------------------------------------
      ! select largest Ritz value and vector
      !----------------------------------------
      jj = MAXLOC(ev(1:kmax))
      imax = jj(1)
      evmax = ev(imax)

      !--------------------------------------------
      ! Ritz vec = sum_{j,kmax} qq(j) * AA(j,imax)
      !--------------------------------------------
      call assign_mult(qq(KDIM+1),qq(1),AA(1,imax))
      do j=2,kmax
        call accum_add_mult(qq(KDIM+1),qq(j),AA(j,imax))
      enddo
      call assign(qq(1),qq(KDIM+1))

!      if (kmax < KDIM) exit

      if (NSTP == istp) then
        !------------------------------------------
        ! eigenvalue upperbound estimate
        !------------------------------------------
        select case(iev)
        case(0)
         if (0==nodeid) write(iout,'("%LANCZOS_EVAL_MAX_EVMAX",2E24.15)')evmax
          eval_max = evmax + ABS(AA(kmax,imax)*TT(kmax+1,kmax))
          dmax     = evmax
          if (ABS(eval_max-dmax)/MIN(ABS(eval_max),ABS(dmax)) > 0.1_DP) then
            if (0==nodeid) then
              write(iout,'("%LANCZOS_EVAL_MAX_RANGE",2E24.15)')dmax,eval_max
            endif
            goto 1000
          endif
        case(1)
         if (0==nodeid) write(iout,'("%LANCZOS_EVAL_MIN_EVMAX",2E24.15)')evmax
          eval_min = 1.0_DP/(evmax + ABS(AA(kmax,imax)*TT(kmax+1,kmax)))
          dmin     = 1.0_DP/ evmax
          if (eval_min < 0.0_DP .or. ABS(dmin-eval_min)/MIN(ABS(eval_min),ABS(dmin)) > 0.1_DP) then
            if (0==nodeid) then
              write(iout,'("%LANCZOS_EVAL_MIN_RANGE",2E24.15)')eval_min,dmin
            endif
            goto 1000
          endif
        end select

      endif

    enddo ! end of do istp
  enddo ! end of do iev

  if (0==nodeid) then
    write(iout,'("%LANCZOS_EVAL_MIN_RANGE",2E24.15)')eval_min,dmin
    write(iout,'("%LANCZOS_EVAL_MAX_RANGE",2E24.15)')dmax,eval_max
  endif

  deallocate(qq,TT,AA,ev,qw)

  call toc(solver_time)
  etime = get_elapse(solver_time)
  if (0 == nodeid) then
    write(iout,'("%GET_EIGEN_BOUND_ETIME_=",E24.15)')etime
  endif

  return
end subroutine


subroutine get_eigen_Qee(phys_param,iout,err,iiter,u,ieo,eval,shift)
!
! Computes maximum eigenvalue lambda for 
!  v 
! (Qee - shift)^(-1) v = lambda v.
!
! and return eval for
! v
! Qee v = eval v  with eval = 1/lambda + shift
!
  use ks_class
  use timer_class
  implicit none
#ifdef _CLOVER
  type(quark_clover), intent(inout) :: phys_param
#else
  type(quark_wilson), intent(inout) :: phys_param
#endif
  integer,               intent(in)    :: iout    ! log file unit
  real(DP),              intent(in)    :: err     ! eigenvalue precison goal
  integer,               intent(inout) :: iiter   ! iteration counter
  type(vfield_gluon_wg), intent(in)    :: u       ! gauge field
  integer,               intent(in)    :: ieo     ! even/odd site index
  real(DP),              intent(out)   :: eval    ! eigenvalue
  real(DP),    optional, intent(in)    :: shift   ! shift value

  character(len=CHARLEN) :: str
  type(field_quark_eo_wg), allocatable :: we(:),ve(:),re
  type(ks_alg), allocatable :: eigen_solver
  type(timer) :: solver_time
  real(DP), allocatable :: sft(:)
  integer :: jiter,nnev,i
  logical :: is_shift
  real(DP) :: tol
  real(DP) :: tol_slv
  integer, parameter :: NEV=4, MEV=20
  integer, parameter :: NSITE=COL*SPIN*NTH*NZ*NY*NX
  real(DP) :: eev(NEV),rtmp,rtmp2,etime
  logical :: eigen_debug

  call new(solver_time)
  call tic(solver_time)

  is_shift = present(shift)
  if (is_shift) then
    allocate(sft(1))
    sft(1) = shift
  endif

  tol     = err            !     eigen solver tolerance
  tol_slv = err*1.0e-1_DP  ! inversion solver tolerance

  allocate(we(1),ve(1))
  call new(we(1),ieo)
  call new(ve(1),ieo)

  allocate(eigen_solver)
  iiter = 0
  write(iout,'("# NEV=",I3)')NEV
#ifdef _DEBUG
  eigen_debug = .TRUE.
#else
  eigen_debug = .FALSE.
#endif
  call new(eigen_solver,n=NSITE,m=MEV,NEV=nev,maxiter=400,tol=tol,debug=eigen_debug,mode=ESLV_MODE_HERMITE)

  do
    call solve(eigen_solver)
    select case(get_status(eigen_solver))
    case (ESLV_OP_NOP)
      cycle
    case (ESLV_OP_DO_MATVEC)
      call unpack(eigen_solver%src_vec,we(1))
      if (is_shift) then
        call assign_inv_mult_Qee_shift(phys_param,iout,tol_slv,jiter,ve,we(1),u,sft)
      else
        call assign_inv_mult_Qee_shift(phys_param,iout,tol_slv,jiter,ve,we(1),u)
      endif
      call   pack(ve(1),eigen_solver%dst_vec)
      iiter = iiter + jiter
    case (ESLV_OP_CONVERGED)
      exit
    case (ESLV_OP_FAIL)
      write(str,'("Error in eigen solver: get_eigen_Qee, file:",A," line:",I5)')__FILE__,__LINE__
      call error_stop(TRIM(str))
    end select
  enddo

  if (0==nodeid) then
    do i=1,NEV
      write(iout,'("%EVS",I3,2E24.15)')i,eigen_solver%eval(i)
    enddo
  endif
  if (is_shift) then
    do i=1,NEV
      eev(i) = 1.0_DP/REAL(eigen_solver%eval(i),kind=DP)+sft(1)
    enddo
    eval = MINVAL(eev(:))
  else
    do i=1,NEV
      eev(i) = 1.0_DP/REAL(eigen_solver%eval(i),kind=DP)
    enddo
    eval = MINVAL(eev(:))
  endif
  if (0==nodeid) then
    do i=1,NEV
      write(iout,'("%EEV",I3,E24.15)')i,eev(i)
    enddo
  endif

#ifdef _DEBUG
!====================================================
! check residual
!====================================================
!       v     v
!  ve = Dee^+ Dee we(1) = Qee we(1)
!
!  re = Qee we(1) - eval we(1)
!
!====================================================
  allocate(re)
  call unpack(eigen_solver%V(:,1),we(1))
  call assign_mult_eoprec_wd(    phys_param,ve(1),we(1),u)
  call assign_mult_eoprec_wd_dag(phys_param,re   ,ve(1),u)
  rtmp = - eval
  call accum_add_mult(re,we(1),rtmp)    ! re = re + we(1)*(-eval)
  rtmp2 = abs2(we(1))
  rtmp = sqrt(abs2(re)/rtmp2/eval**2)
  deallocate(re)
  if (0 == nodeid) then
    write(*,'("%REL_ERR_=",E24.15)')rtmp
    write(*,'("%EVEC2_=",E24.15)')rtmp2
  endif
!====================================================
#endif

  call delete(eigen_solver)
  deallocate(eigen_solver)
  deallocate(ve,we)
  if (is_shift) deallocate(sft)

  call toc(solver_time)

  etime = get_elapse(solver_time)

  if (0 == nodeid) then
    write(iout,'("%GET_EIGEN_ETIME_=",E24.16)')etime
  endif

  return
end subroutine


subroutine assign_mult_function_Qee(this,itype,iout,rye,ye,u)
!                              v
!= Multiply matrix function of Qee using Rational approximation
!
!  rye = R(Qee) ye
!
!  R(x) = sqrt(x)     : itype = +1
!  or   = 1/sqrt(x)   : itype = -1
!  or   = x**(1/4)    : itype = +2
!  or   = x**(-1/4)   : itype = -2
!
!        v          v
!  Qee = Dee^{\dag} Dee
!
! This function is used in the Hamiltonian computation.
!
  use pfcoef_class
  implicit none
  type(hmc_ratio_quark_wilson), intent(inout) :: this
  integer,                      intent(in)    :: itype  ! function type index
  integer,                      intent(in)    :: iout   ! logfile unit
  type(field_quark_eo_wg),      intent(inout) :: rye
  type(field_quark_eo_wg),      intent(in)    ::  ye
  type(vfield_gluon_wg),        intent(in)    ::   u

  type(partial_fraction_coef) :: pfcoef
  type(field_quark_eo_wg), allocatable :: we(:)
  real(DP) :: factor,a,b,err
  integer :: ir,irmin,irmax,iiter

  if (allocated(this%alp)) deallocate(this%alp)
  if (allocated(this%bet)) deallocate(this%bet)

  factor = this%approx_range_factor_hamil
  a = get_min(this%eigen_func_eval)/factor
  b = get_max(this%eigen_func_eval)*factor
  irmax = 64
  irmin = 6
  do ir=irmin,irmax,2
    call new(pfcoef,itype,ir,a,b)
    if (pfcoef%minimaxerror < this%tol_approx_hamil) then
      this%nratio = ir
      allocate(this%alp(0:this%nratio))
      allocate(this%bet(1:this%nratio))
      this%alp(:) = pfcoef%alp(:)
      this%bet(:) = pfcoef%bet(:)
      if (0 == nodeid) then
        write(iout,'("# order=",I3," minmaxerr=",E24.15," [a,b]=",2E24.15)')ir,pfcoef%minimaxerror, &
 &                                                                          a,b
      endif
      call delete(pfcoef)
      exit
    endif
    call delete(pfcoef)
  enddo

  !======================================================
  ! rye = alp(0) ye + sum_{j=1,n} alp(j)*(Q - bet(j)) \ ye
  !======================================================
  allocate(we(1:this%nratio))
  err = this%hamil_inv%tol
  iiter = 0
  call assign_inv_mult_Qee_shift(this%phys_param,iout,this%hamil_inv%tol,iiter,we,ye,u,this%bet)
  call inc(this%hamil_inv,iiter)
  call assign_mult(rye,we(1),this%alp(1))
  do ir=2,this%nratio
    call accum_add_mult(rye,we(ir),this%alp(ir))
  enddo
  call accum_add_mult(rye,ye,this%alp(0))

  deallocate(we)
  if (allocated(this%alp)) deallocate(this%alp)
  if (allocated(this%bet)) deallocate(this%bet)

  return
end subroutine

subroutine check_rational_approx(this,iout,u)
!
! check rational approximation function
!
! This routine is used for debugging.
!
  implicit none
  type(hmc_ratio_quark_wilson), intent(inout) :: this
  integer,                      intent(in)    :: iout
  type(vfield_gluon_wg),        intent(in)    ::  u
  type(field_quark_eo_wg), allocatable :: we,ve,re
  real(DP) :: rtmp0,rtmp1,rtmp2
  integer :: ieo

  allocate(we,ve,re)
  do ieo=0,1 
    call new(we,ieo)
    call new(ve,ieo)
    call new(re,ieo)

    call set_gaussian_noise(we)

    !=================================================
    !  re = [Qee^(-1/2)]^2 we
    !=================================================
    call assign_mult_function_Qee(this,TYPE_INV_SQUARE_ROOT,iout,ve,we,u)
    call assign_mult_function_Qee(this,TYPE_INV_SQUARE_ROOT,iout,re,ve,u)
    !=================================================
    !  ve = Qee re  = we 
    !=================================================
    call assign_mult_Qee(this%phys_param,ve,re,u)
    !=================================================
    !  ve = ve - we
    !=================================================
    call accum_sub(ve,we)
    rtmp0 = abs2(we)
    rtmp1 = abs2(ve)
    rtmp2 = sqrt(rtmp1/rtmp0)
    write(*,'("|[ Qee (Qee^(-1/2))^2 - 1 ]we|/|we|=",E24.15)')rtmp2

    !=================================================
    !  re = [Qee^(1/2)]^2 re  = we 
    !=================================================
    call assign_mult_function_Qee(this,TYPE_SQUARE_ROOT,iout,ve,re,u)
    call assign_mult_function_Qee(this,TYPE_SQUARE_ROOT,iout,re,ve,u)
    !=================================================
    !  re = re - we
    !=================================================
    call accum_sub(re,we)
    rtmp0 = abs2(we)
    rtmp1 = abs2(re)
    rtmp2 = sqrt(rtmp1/rtmp0)
    write(*,'("|[ (Qee^(1/2))^2 (Qee^(-1/2))^2 - 1 ]we|/|we|=",E24.15)')rtmp2


    !=================================================
    !  re = [Qee^(-1/4)]^4 we
    !=================================================
    call assign_mult_function_Qee(this,TYPE_INV_QUAD_ROOT,iout,ve,we,u)
    call assign_mult_function_Qee(this,TYPE_INV_QUAD_ROOT,iout,re,ve,u)
    call assign_mult_function_Qee(this,TYPE_INV_QUAD_ROOT,iout,ve,re,u)
    call assign_mult_function_Qee(this,TYPE_INV_QUAD_ROOT,iout,re,ve,u)
    !=================================================
    !  ve = Qee re  = we 
    !=================================================
    call assign_mult_Qee(this%phys_param,ve,re,u)
    !=================================================
    !  ve = ve - we
    !=================================================
    call accum_sub(ve,we)
    rtmp0 = abs2(we)
    rtmp1 = abs2(ve)
    rtmp2 = sqrt(rtmp1/rtmp0)
    write(*,'("|[ Qee (Qee^(-1/4))^4 - 1 ]we|/|we|=",E24.15)')rtmp2

    !=================================================
    !  re = [Qee^(1/4)]^4 re  = we 
    !=================================================
    call assign_mult_function_Qee(this,TYPE_QUAD_ROOT,iout,ve,re,u)
    call assign_mult_function_Qee(this,TYPE_QUAD_ROOT,iout,re,ve,u)
    call assign_mult_function_Qee(this,TYPE_QUAD_ROOT,iout,ve,re,u)
    call assign_mult_function_Qee(this,TYPE_QUAD_ROOT,iout,re,ve,u)
    !=================================================
    !  re = re - we
    !=================================================
    call accum_sub(re,we)
    rtmp0 = abs2(we)
    rtmp1 = abs2(re)
    rtmp2 = sqrt(rtmp1/rtmp0)
    write(*,'("|[ (Qee^(1/4))^4 (Qee^(-1/4))^4 - 1 ]we|/|we|=",E24.15)')rtmp2

  enddo
  deallocate(we,ve,re)

  return
end subroutine

subroutine get_initial_guess_multi_shift(phys_param,x1,k,b,x0,shift)
!
! Compute initial guess for
!    v
!  ( Qee^2 - shift(k)(gamma5/Fee)^2 ) x0 = b
!
! using previous soutions from
!
!    v
!  ( Qee^2 - shift(j)(gamma5/Fee)^2 ) x1(j) = b,    j=k+1,k+2,...,Np-1,Np
!
! via
!
!  x0 = \sum_j={k+1}^{N} x1(j) c(j)
!
! where c(j) is determined with linear minimization (QR method).
!
!
! This routine is used with pure CG algorithm for shifted systems.
!
  implicit none
#ifdef _CLOVER
  type(quark_clover), intent(inout) :: phys_param
#else
  type(quark_wilson), intent(inout) :: phys_param
#endif
  type(field_quark_eo_wg), intent(inout) :: x1        ! initial geuss for shift(k) system
  integer,                 intent(in)    :: k         ! shift number
  type(field_quark_eo_wg), intent(in)    :: b         ! right hand side vector for shifted equations
  type(field_quark_eo_wg), intent(in)    :: x0(:)     ! previous-shifted solutions
  real(DP),                intent(in)    :: shift(:)  ! shift values

  type(field_quark_eo_wg), allocatable   :: V(:)
  complex(DP),             allocatable   :: R(:,:)
  real(DP),                allocatable   :: delta(:)
  integer :: i,j,nn,n,np
  complex(DP) :: ctmp,ctmp2
  real(DP)    :: rtmp,rtmp2,res

#ifdef _BACKWARD_
  np = SIZE(x0)
  nn = np - k + 1
  n  = nn - 1
  allocate(V(1:nn),R(1:nn,1:nn),delta(1:n))
  do i=1,n
    delta(i) = shift(k+i) - shift(k)
  enddo
#else
  nn = k
  n  = k - 1 
  allocate(V(1:nn),R(1:nn,1:nn),delta(1:n))
  do i=1,n
    delta(i) = shift(k-i) - shift(k)
  enddo
#endif

  !=======================================================
  !  v(i) = (Qee - shift(k  )) x0(k+i)
  !       = (Qee -(shift(k+i) + shift(k+i) - shift(k))) x0(k+i)
  !       = (Qee - shift(k+i)) x0(k+i) + x0(k+i) * delta(i)
  !       = b + x0(k+i) * delta(i)
  !=======================================================
  do i=1,n
    !----------------------------------------------
    ! v(i) = x0(k+i)*delta(i) + b
    !----------------------------------------------
#ifdef _BACKWARD_
    call assign_mult_add(V(i),x0(k+i),delta(i),b)
#else
    call assign_mult_add(V(i),x0(k-i),delta(i),b)
#endif
  enddo
  call assign(V(nn),b)

  !=======================================================
  ! QR decomposition for V = (v(1),...,,v(k+1))
  !  V = QR, V := Q
  !=======================================================
  R(:,:) = Z0
  do i=1,nn
    do j=1,i-1
      ctmp = prod(V(j),V(i))
      ctmp2 = -ctmp
      call accum_add_mult(V(i),V(j),ctmp2)
      R(j,i) = ctmp
    enddo
    do j=1,i-1
      ctmp = prod(V(j),V(i))
      ctmp2 = -ctmp
      call accum_add_mult(V(i),V(j),ctmp2)
      R(j,i) = R(j,i) + ctmp
    enddo
    rtmp = SQRT(abs2(V(i)))
    R(i,i) = rtmp
    rtmp = 1.0_DP/rtmp
    call accum_mult(V(i),rtmp)
  enddo

#ifdef _GGG_
!==========
do j=1,nn
do i=1,nn
  ctmp = prod(V(i),V(j))
  write(*,'("Q^dagQ=",2I3,2E24.15)')i,j,ctmp
enddo
enddo
!==========
#endif

#ifdef _GGG_
  !-------------------------------------
  ! normalized residual |b-Ax|/|b|
  !-------------------------------------
  rtmp = abs2(b)
  res = REAL(R(nn,nn),kind=DP)/SQRT(rtmp)
  write(*,'("%MULTI_SHIFT_CHRON_ERR_=",2E24.15)')res
#endif

  !-------------------------------------
  ! R(1:n,nn) <= R(1:n,1:n) \ R(1:n,nn)
  !-------------------------------------
  do i=n,1,-1
    do j=i+1,n
      R(i,nn) = R(i,nn) - R(i,j)*R(j,nn)
    enddo
    rtmp = REAL(R(i,i),kind=DP)
    R(i,nn) = R(i,nn) / rtmp
  enddo

  !-------------------------------------
  ! x1 = sum_{i=k+1,np}[x0(i)*R(i,nn)]
  !-------------------------------------
#ifdef _BACKWARD_
  call assign_mult(x1,x0(k+1),R(1,nn))
  do i=2,n
    call accum_add_mult(x1,x0(k+i),R(i,nn))
  enddo
#else
  call assign_mult(x1,x0(k-1),R(1,nn))
  do i=2,n
    call accum_add_mult(x1,x0(k-i),R(i,nn))
  enddo
#endif

  deallocate(V,R,delta)
  return
end subroutine

#ifdef _GPU_SOLVER_
subroutine v3_assign_inv_mult_Qee_shift(phys_param,iout,tol,iiter,Aye,ye,u,shift)
!                 
! Multiply inverse (Qee - shift_j) on a vector ye using normal CG algorithm.
!
!  Aye(j) = (Qee - shift_j) \ ye
!
! where
!        v       v
!  Qee = Dee^dag Dee
! 
! 
  use gpu_block_solver_class
  use solver_class
  use timer_class
  implicit none
#ifdef _CLOVER
  type(quark_clover), intent(inout) :: phys_param
#else
  type(quark_wilson), intent(inout) :: phys_param
#endif
  integer,  intent(in) :: iout
  real(DP), intent(in) :: tol
  integer,  intent(inout) :: iiter
  type(field_quark_eo_wg), intent(inout) :: Aye(:)   ! even/odd site fermion field block (output)
  type(field_quark_eo_wg), intent(in)    ::  ye      ! even/odd site fermion field (input)
  type(vfield_gluon_wg),   intent(in)    ::   u      ! gauge field
  real(DP), optional,      intent(in)    :: shift(:) ! shift values

  integer, parameter :: MAXITER=5000
  type(gpu_block_solver) :: gpu_precnd

  type(field_quark_eo_wg), allocatable :: we,ve(:),re(:)
  type(counter) :: imult
  type(timer) :: solver_time
  character(len=CHARLEN) :: str
  integer, parameter :: NSITES=COL*SPIN*NTH*NZ*NY*NX
  integer  :: ir,nr
  integer  :: mult_count
  integer  :: istat,isite,iter
  logical  :: is_shift
  real(DP), allocatable :: tot_err(:)
  real(DP) :: etime,rtmp
  real(DP) :: bsrc,rnorm,err,kappa
  logical :: is_debug

  call new(solver_time)
  call tic(solver_time)

#ifdef _DEBUG
  is_debug = .TRUE.
#else
  is_debug = .FALSE.
#endif

  is_shift = present(shift)
  nr = SIZE(Aye(:))

  if (is_shift) then
    if (SIZE(shift(:)) /= nr) then
      write(str,'("Error: SIZE(shift) /= SIZE(Aye): v3_assign_inv_mult_Qee_shift:",A,I4)')__FILE__,__LINE__
      call error_stop(TRIM(str))
    endif
  endif

  imult = mult_iter

  gpu_precnd%num_pf =  nr
  gpu_precnd%tol    = tol
#ifdef _CLOVER
  call new(gpu_precnd,u,ye%ieo,get_ptr_inverse_clover_term(phys_param))
#else
  call new(gpu_precnd,u,ye%ieo)
#endif

  allocate(we,ve(nr),re(nr),tot_err(nr))
  call new(we,ye%ieo)
  do ir=1,nr
    call new(Aye(ir),ye%ieo)
    call new( ve(ir),ye%ieo)
    call new( re(ir),ye%ieo)
    call clear(Aye(ir))
  enddo

  kappa = get_kappa(phys_param)

  bsrc = sqrt(abs2(ye))
  !------------------
  ! residual r = y
  !------------------
  do ir=1,nr
    call assign(re(ir),ye)
  enddo
  err = 1.0_DP
  tot_err(:) = err
  iiter = 0
  iter = 0
  if (is_debug) then
    do ir=1,nr
      mult_count = get_count(mult_iter)-get_count(imult)
      if (nodeid==0) write(*   ,'("#",I5," ERR_=",E24.15," MULT_=",I10," NPF_=",I4)') &
 &                          iter/nr,tot_err(ir),get_count(mult_iter)-get_count(imult),ir
    enddo
  endif


  do while (iter < MAXITER*nr)
    iiter = iiter + 1 

    !=========================
    ! V = M R
    !=========================
    gpu_precnd%err_norm = err
    if (is_shift) then
      call assign_inv_mult_Qee_shift(gpu_precnd,kappa,ve,re,shift)
    else
      call assign_inv_mult_Qee_shift(gpu_precnd,kappa,ve,re)
    endif

    !=========================
    ! X = X + V
    ! R = R - W
    !=========================
    iter = iter + gpu_precnd%iter
    do ir=1,nr
      !-------------------------------
      ! w(j) = (Qee - shift(j)) v(j)
      !-------------------------------
      call assign_mult_Qee(phys_param,we,ve(ir),u)
      if (is_shift) then
        rtmp = -shift(ir)
        call accum_add_mult(we,ve(ir),rtmp)
      endif
      call accum_add(Aye(ir),ve(ir))
      call accum_sub( re(ir),we)
      rtmp = sqrt(abs2(re(ir)))/bsrc
      tot_err(ir) = rtmp
    enddo

    err = MAXVAL(tot_err)
    if (is_debug) then
      do ir = 1,nr
        if (nodeid==0) write(*   ,'("#",I5," ERR_=",E24.15," MULT_=",I10," NPF_=",I4)') &
 &                           iter/nr,tot_err(ir),get_count(mult_iter)-get_count(imult),ir
      enddo
    endif
    if ( err < tol ) goto 100

  enddo

  deallocate(we,ve,re,tot_err)
  call delete(gpu_precnd)

  call error_stop("# RHMC CG iteration does not converge.")

100 continue
  if (nodeid==0) write(iout,'("# Residual iteration (+CGGPU) converged.")')
  do ir = 1,nr
    if ( ir < nr ) then
      if (nodeid==0) write(iout,'("#",I5," ERR_=",E24.15," MULT_=",I10," NPF_=",I4)') &
 &                        iter/nr,tot_err(ir),get_count(mult_iter)-get_count(imult),ir
    else
      call toc(solver_time)
      etime = get_elapse(solver_time)
      if (nodeid==0) write(iout,'("#",I5," ERR_=",E24.15," MULT_=",I10," NPF_=",I4," ETIME_=",E24.15)') &
 &                        iter/nr,tot_err(ir),get_count(mult_iter)-get_count(imult),ir,etime
    endif
  enddo

  deallocate(we,ve,re,tot_err)
  call delete(gpu_precnd)

  return
end subroutine
#endif

end module
